@@ -83,6 +83,7 @@ typedef enum
   DK_STAT_ARGS, /* Arguments are not usable. */
   DK_STAT_REVOKED,    /* Key has been revoked. */
   DK_STAT_INTERNAL, /* cannot call this routine in this context.  Internal error. */
+  DK_STAT_GRANULARITY, /* Granularity mismatch: sender doesn't match g= option. */
 } DK_STAT;
 
 typedef enum
@@ -134,6 +135,7 @@ typedef struct
   char *domain;     /* d= */
   char *selector;   /* s= */
   char *signature;    /* b= */
+  char *granularity;    /* g= */
   char *keyrec; /* caller supplied TXT key record */
   char *policyrec; /* caller supplied TXT policy record */
   int canon;      /* c= canonicalization being applied. */
@@ -169,6 +171,7 @@ static char *errors[] =
   "DK_STAT_ARGS: Arguments are not usable.",
   "DK_STAT_REVOKED: Key has been revoked.",
   "DK_STAT_INTERNAL: cannot call this routine in this context.  Internal error.",
+  "DK_STAT_GRANULARITY: Granularity mismatch: sender doesn't match g= option."
 };
 /* STOPPRIV */
 /* STOPHEAD */
@@ -265,6 +268,7 @@ static DK_STAT dkinit_new_header(DK *dk)
   {
     return DKERR(DK_STAT_NORESOURCE);
   }
+  memset(dk->header,0,DK_BLOCK);
   dk->headerlen = 0;
   dk->headerlinelen = 1;  /* always store the first char. */
   dk->in_headers = 1;
@@ -273,6 +277,9 @@ static DK_STAT dkinit_new_header(DK *dk)
   dk->sender = NULL;
   dk->dksign = NULL;
   dk->dktrace = NULL;
+  dk->domain = NULL;
+  dk->selector = NULL;
+  dk->signature = NULL;
   dk->canon = DK_CANON_SIMPLE;
   dk->state = 0;
   dk->headers = 0;
@@ -281,6 +288,7 @@ static DK_STAT dkinit_new_header(DK *dk)
   dk->last_char = '\0';
   dk->trace = NULL;
   dk->traceheader = NULL;
+  dk->granularity = NULL;
   dk->keyrec = NULL;
   dk->policyrec = NULL;
 #ifdef DK_HASH_BUFF
@@ -289,6 +297,7 @@ static DK_STAT dkinit_new_header(DK *dk)
   {
     return DKERR(DK_STAT_NORESOURCE);
   }
+  memset(dk->hash_buff,0,DK_BLOCK);
   dk->hash_buff_len = 0;
 #endif
   return DKERR(DK_STAT_OK);
@@ -568,9 +577,8 @@ static DK_STAT dkparse822(char *address, unsigned int offset)
     else if (*from == '"')
     {
       /* we have to ignore parenthesis inside quotes. */
-      level = 1;
       from++;
-      while (level)
+      for (;;)
       {
         if (*from == '\0')
         {
@@ -578,7 +586,8 @@ static DK_STAT dkparse822(char *address, unsigned int offset)
         }
         else if (*from == '"')
         {
-          --level;
+          from++;
+          break;
         }
         else if (*from == '\\' && from[1])
         {
@@ -1162,15 +1171,15 @@ static DK_STAT dkheaders(DK *dk)
             dk->sender = 0;
             return DKERR(DK_STAT_SYNTAX);
           }
-          else if (!dk->sender_beforesign)
+        }
+        else if (!dk->sender_beforesign)
+        {
+          dk->sender_beforesign = dk_strdup(dk->header + i);
+          if (dkparse822(dk->sender_beforesign, 7) != DK_STAT_OK)
           {
-            dk->sender_beforesign = dk_strdup(dk->header + i);
-            if (dkparse822(dk->sender_beforesign, 7) != DK_STAT_OK)
-            {
-              DK_MFREE(dk->sender_beforesign);
-              dk->sender_beforesign = 0;
-              return DKERR(DK_STAT_SYNTAX);
-            }
+            DK_MFREE(dk->sender_beforesign);
+            dk->sender_beforesign = 0;
+            return DKERR(DK_STAT_SYNTAX);
           }
         }
       }
@@ -1287,91 +1296,122 @@ static DK_STAT dkheaders(DK *dk)
  * ptr may be NULL, in which case the bytes are just counted, not copied.
  * Feel free to call this twice; once to get the length, and again to
  * copy the data.
- * If we preserve duplicate headers, actually return something (len)
- * If we remove duplicate headers, len may be innacurate (greater than what's
- * really needed when calling with ptr == NULL
+ * NOTE: If the return value is 0 then an error occured.
+ *	It's a good idea to check for this
  */
 int dk_headers(DK *dk, char *ptr)
 {
-  int len;
-  char *headers;
-  int k, m;
-  if (!dk)
+  register int len = 0;
+  register int label_len = 0;
+  int list_size;
+  char *header_start = NULL;
+  char *header_end = NULL;
+  char *h_label = NULL;
+  char *h_list = NULL;
+
+  char *dupe_list = NULL;
+
+
+  if (ptr) //if you dont check if return is 0, ptr can be anything
+  	ptr[0] = 0;
+
+  if ((!dk)||(dk->dkmarker != DKMARK))
   {
     return 0;
   }
-  if (dk->dkmarker != DKMARK)
-  {
+  list_size = dk->headermax;
+
+  h_list = (char *) DK_MALLOC(list_size);
+  if (!h_list)
     return 0;
-  }
+  memset(h_list,0,list_size);
+  h_list[len++] = ':';
 
-  headers = dk->header + dk->start_signed;
+  header_start = dk->header + dk->start_signed;
 
-  len = 0;
-  m = dk->start_signed;
-  for (k = dk->start_signed; k < dk->headerlen; k++)
+  for (header_end = header_start; header_end < (dk->header + dk->headerlen); header_end++)
   {
-    if (dk->header[k] == '\0')
+    if (*header_end == '\0')
     {
-      m = k+1;
+      header_start = header_end+1;//point to start of next line
     }
-    else if (dk->header[k] == ':' && m >= 0)
+    else if (*header_end == ':' && header_start)
     {
-      if (ptr)
-      {
-        memcpy(ptr + len, dk->header + m, k - m + 1);
-      }
-      len += k - m + 1;
-      m = -1;
+      label_len = (header_end - header_start) + 1;
+	  //grow list array
+	  if ((len+label_len) >= list_size)
+	  {
+	  	char *temp = NULL;
+		list_size += dk->headermax;
+		temp = DK_MALLOC(list_size);
+		if (!temp)
+		{
+			len = 1; //out of memory?
+			goto END;
+		}
+		memset(temp,0,list_size);
+		memcpy(temp,h_list,len);
+		DK_MFREE(h_list);
+		h_list = temp;
+	  }
+
+      memcpy(h_list + len, header_start, label_len);
+      len += label_len;
+      header_start = NULL;
     }
   }
 
-  if ((dk->opts & DKOPT_RDUPE)&&(ptr))
+  if (dk->opts & DKOPT_RDUPE)
   {
-    int headpos = 0;
-    char * copy = (char*)DK_MALLOC(len + 2);
-    copy[0] = ':';
-    memcpy(copy + 1, ptr, len-1);
-    copy[len] = ':';
-    copy[len + 1] = 0;
-    ptr[0] = 0;
-    //fprintf(stderr,"!!%s\n",copy);
-    m = 0;
-    for (k = 1; k < len + 2; k++)
-    {
-      if (copy[k] == ':')
-      {
-        int found = 0;
-        int sub;
-        for (sub = 0; sub < len + 2; sub++)
-        {
-          if ((copy[sub] == ':') && (sub != m))
-          {
-            if (!strncmp(copy + sub, copy + m, k - m))
-            {
-              found = 1;
-              break;
-            }
-          }
-        }
-        if (!found)
-        {
-          memcpy(ptr + headpos, copy + m + 1, k - m);
-          headpos += k - m;
-          ptr[headpos] = 0;
-        }
-        m = k;
-      }
-    }
-    DK_MFREE(copy);
-    len = headpos;
+
+    h_label = (char *) DK_MALLOC(list_size);
+	dupe_list = (char *) DK_MALLOC(list_size);
+
+    if (!h_label || !dupe_list)
+	{
+      len = 1;
+	  goto END;
+	}
+    memset(h_label,0,list_size);
+    memset(dupe_list,0,list_size);
+
+    len = 0;
+  	header_start = h_list;
+	for (header_end = header_start + 1; *header_end != '\0'; header_end++)
+	{
+		if (*header_end == ':')
+		{
+			label_len = (header_end - header_start)+1;
+	        memcpy(h_label, header_start, label_len);
+    	    h_label[label_len] = 0;
+			if ((strstr(h_list,h_label) == header_start)  && (strstr(header_start + 1,h_label) == NULL) )
+			{
+				memcpy(dupe_list + len ,h_label,label_len);
+				len += label_len -1;
+			}
+			header_start = header_end;
+		}
+	}
+	len++;
+	if (ptr)
+		memcpy(h_list, dupe_list, len);
   }
 
-  if (len && ptr)
+  if ((ptr != NULL)&&(len > 1))
   {
-    ptr[len-1] = '\0';  /* change the last colon into a null. */
+    memcpy(ptr,h_list+1,(len-2));//dont count the prefix and postfix'd ":"
+    ptr[len-2] = 0;
   }
-  return len; //return size including '\0'
+  END:
+  if (h_label)
+    DK_MFREE(h_label);
+  if (h_list)
+    DK_MFREE(h_list);
+  if (dupe_list)
+    DK_MFREE(dupe_list);
+
+  return --len;
+
 }
 
 /* HEADER */
@@ -1631,6 +1671,25 @@ char *dk_address(DK *dk)
   return "N";     /* TC14-5 */
 }
 
+/* HEADER */
+/*
+ * Returns a pointer to a null-terminated string containing the granularity
+ * value found in the selector DNS record, if any, but only after dk_end
+ * has been called. Otherwise returns NULL.
+ */
+char *dk_granularity(DK *dk)
+{
+  if (!dk)
+  {
+    return NULL;
+  }
+  if (dk->dkmarker != DKMARK)
+  {
+    return NULL;
+  }
+  return dk->granularity;
+}
+
 /* HEADER */
 /*
  * Called at end-of-message (before response to DATA-dot, if synchronous with SMTP session).
@@ -1711,12 +1770,13 @@ DK_STAT dk_end(DK *dk, DK_FLAGS *dkf)
   case DK_SIGNING_NOVERIFY:
     {
       unsigned char md_value[1024];
-      int md_len,i;
-      char *s = NULL;
+      int md_len=0;
+      char *s = NULL; //pointer to domain portion of email addy
+      char *sndr = NULL;
       char *domainkeys; /* malloc'ed */
       char *txtrec; /* malloc'ed */
       char *pubkeyvals[4];
-      int okay;
+      int i;
       DK_STAT st;
 
       BIO *bio, *b64;
@@ -1732,40 +1792,47 @@ DK_STAT dk_end(DK *dk, DK_FLAGS *dkf)
       /* Or is a subdomain of the d data.  TC26-1, etc. */
       if (dk->sender)
       {
-        s = dk->sender;
+        sndr = dk->sender;
       }
       else if (dk->sender_beforesign && (dk->signing == DK_SIGNING_NOVERIFY))
       {
-        s = dk->sender_beforesign;
+        sndr = dk->sender_beforesign;
       }
       else
       {
-        s = dk->from;
+        sndr = dk->from;
       }
 
-      if (!s)
+      if (!sndr)
       {
         return DKERR(DK_STAT_SYNTAX);
       }
-      /* we already know that s has an '@' in it; otherwise it would not
+      /* we already know that sndr has an '@' in it; otherwise it would not
       * have passed dkparse822. */
-      s = strchr(s, '@')+1;
+      
+      /* sndr can also be a one-character string if the from/sender header
+      * was there but empty */
+      if (!(s = strchr(sndr, '@')))
+      {
+        return DKERR(DK_STAT_SYNTAX);
+      }
+      s++; // we skip the tag
 
-      i = 0;
+      i = (int)strlen(s);
       if (strcasecmp(dk->domain, s))
       {
-        for (; i < (int)strlen(s); i++)
+        for (; md_len < i; md_len++)
         {
-          if (s[i] == '.')
+          if (s[md_len] == '.')
           {
-            if (!strcasecmp(dk->domain, s + i + 1))
+            if (!strcasecmp(dk->domain, s + md_len + 1))
             {
               break;
             }
           }
         }
       }
-      if (i == (int)strlen(s))
+      if (md_len == i)
       {
         return DKERR(DK_STAT_SYNTAX);
       }
@@ -1832,7 +1899,19 @@ DK_STAT dk_end(DK *dk, DK_FLAGS *dkf)
         }
         *dkf |= DK_FLAG_SET;
       }
-
+      /* Store the granularity, if any, and check it against the local part
+       * of the sending address */
+ 
+      if (pubkeyvals[3] && *pubkeyvals[3])
+      {
+        dk->granularity = dk_strdup(pubkeyvals[3]);
+		// Note that sndr has a leading 'F' or 'S' 
+        if (strncasecmp(dk->granularity, sndr+1, strcspn(sndr+1,"@")))
+		{
+	        DK_MFREE(txtrec);
+	  		return DKERR(DK_STAT_GRANULARITY);
+		}
+      }
       if (!pubkeyvals[0])
       {
         DK_MFREE(txtrec);
@@ -1870,9 +1949,9 @@ DK_STAT dk_end(DK *dk, DK_FLAGS *dkf)
       }
 
       /* using that key, verify that the digest is properly signed */
-      okay = EVP_VerifyFinal(&dk->mdctx, md_value, md_len, publickey);
+      i = EVP_VerifyFinal(&dk->mdctx, md_value, md_len, publickey);
 
-      if (okay > 0)
+      if (i > 0)
       {
         st = DK_STAT_OK;
       }
@@ -2066,6 +2145,10 @@ DK_STAT dk_free(DK *dk, int doClearErrState)
   {
     DK_MFREE(dk->policyrec);
   }
+  if (dk->granularity)
+  {
+    DK_MFREE(dk->granularity);
+  }
 #ifdef DK_HASH_BUFF
   DK_MFREE(dk->hash_buff);
 #endif
@@ -2100,3 +2183,5 @@ const char *DK_STAT_to_string(DK_STAT st)
     return errors[st];
   }
 }
+
+
